"""
-------------------------------------------------------------------------------

Analog damped-oscillator filter-element design

-------------------------------------------------------------------------------
"""

import numpy as np
import scipy.special

from irides.resources.containers import points


# noinspection SpellCheckingInspection
def unit_peak_times(cartesian_poles: np.ndarray) -> np.ndarray:
    """Returns peak times for a unit-scale impulse response.

    The timepoint where h(t) peaks, with tau = 1, is

              1                             1
        t* = ---- theta csc(theta) = ----------------
              wn                      wn sinc(theta)

    where theta is angle of the reference pole to the
    negative real axis, and wn is the natural frequency.

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.

    Returns
    -------
    np.ndarray
        Unit peak time values.
    """

    # extract wn, theta
    wn, theta = extract_wn_and_theta(cartesian_poles)

    # calc and return
    return 1.0 / (wn * np.sinc(theta / np.pi))


# noinspection SpellCheckingInspection
def unit_peak_values(cartesian_poles: np.ndarray) -> np.ndarray:
    """Returns peak value for unit-scale impulse response.

    The value at h(t*) is the peak value. For unit tau, the
    peak value is

        h(t*) = wn exp(-theta cot(theta))
              = wn exp(-cos(theta) / sinc(theta))

    where theta is the angle of the reference pole to the
    real axis, and wn is the natural frequency.

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.

    Returns
    -------
    np.ndarray
        Unit peak values.
    """

    # extract wn, theta
    wn, theta = extract_wn_and_theta(cartesian_poles)

    # calc and return
    return wn * np.exp(-np.cos(theta) / np.sinc(theta / np.pi))


# noinspection SpellCheckingInspection
def peak_value_coordinates(cartesian_poles: np.ndarray, tau=1.0):
    """Returns (time, value) coordinates for a scaled impulse response.

    Provided that there is a non-zero real part to the reference
    poles, there will be a single peak of the impulse response h(t).

                  ....x...  <---- h(t*)
              ....        ....
           ...                ..
         ..                     ...
        .                          ....
    ----|-------------|-------------------------> t
        0
                     t*

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.
    tau: float
        The impulse-response scale factor (default = 1.0).

    Returns
    -------
    points.ContinuousTimePoints
        Scaled peak coordinates.
    """

    # put together into a point coordinate and return
    return points.ContinuousTimePoints(
        unit_peak_times(cartesian_poles) * tau,
        unit_peak_values(cartesian_poles) / tau,
    )


# noinspection SpellCheckingInspection
def nth_null_times(cartesian_poles: np.ndarray, n=1, tau=1.0) -> np.ndarray:
    """Returns times of the nth zero crossing of the impulse response.

    The null locations are purely governed by the sinusoidal component
    of h(t). Nulls exist only when |theta| > 0, where /theta/ is the
    polar angle of the reference pole.

    The temporally scaled nth null time is

                     n pi
        t_null(n) = ------ csc(|theta|) tau,
                      wn

    where /wn/ and /theta/ are the polar coordinates of /reference_pole/.

    In the case that /theta/ == 0, t_null = inf.

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.
    n: int
        The nth zero crossing (starts with 1).
    tau: float
        The impulse-response scale factor (default = 1.0).

    Returns
    -------
    np.ndarray
        nth null times of a scaled impulse response.
    """

    # extract wn, theta
    wn, theta = extract_wn_and_theta(cartesian_poles)

    # calc and return
    with np.errstate(divide="ignore"):

        return (tau * n * np.pi / wn) / np.sin(np.abs(theta))


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, dt=0.0):
    """Returns a function to calculate the requested moment.

    Returns an anonymous function with arguments (tau, wn, theta), where
    wn is the natural frequency of the pole pair and theta is the inclination
    angle from the negative real axis. The returned function captures the
    dt argument in this calling function in order to make a discretization
    adjustment to the moments.

    Note that the generated functions are not purely continuous time.
    Instead, they bridge continuous and discrete time by applying order-dt
    corrections to the moments. This is required because the CT series are
    in fact implemented in DT.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with arguments (tau, wn, theta) as np.ndarrays.
    """

    # generator defs with dt capture
    # noinspection PyUnusedLocal
    def m0_gen(
        wn: np.ndarray, theta: np.ndarray, tau: np.ndarray
    ) -> np.ndarray:
        return np.ones_like(wn)

    # noinspection PyUnusedLocal
    def m1_gen(
        wn: np.ndarray, theta: np.ndarray, tau: np.ndarray
    ) -> np.ndarray:
        return 2.0 * tau * np.cos(theta) / wn

    # noinspection PyUnusedLocal
    def m2_gen(
        wn: np.ndarray, theta: np.ndarray, tau: np.ndarray
    ) -> np.ndarray:
        return 2.0 * (1.0 + 2.0 * np.cos(2.0 * theta)) / np.power(wn / tau, 2)

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection,PyUnusedLocal
def full_width_generator(dt=0.0):
    """
    Returns anonymous function for the full-width. `dt` is captured here,
    and the anonymous function has arguments (wn, theta).

    Parameters
    ----------
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with arguments (tau, wn, theta).
    """

    # noinspection PyPep8Naming,PyUnusedLocal
    def fw_gen(
        wn: np.ndarray, theta: np.ndarray, tau: np.ndarray
    ) -> np.ndarray:
        """
        Full width of the damped oscillator is only defined for
        -pi/4 < theta < pi/4  (theta from the negative real axis).
        Returns np.nan otherwise.
        """

        with np.errstate(invalid="ignore"):

            M1 = 2.0 * np.cos(theta) / (wn / tau)
            RHW = np.sqrt(np.cos(2.0 * theta) / (1.0 + np.cos(2.0 * theta)))
            return 2.0 * M1 * RHW

    return fw_gen


# noinspection SpellCheckingInspection
def key_spectral_features(cartesian_poles: np.ndarray, tau=1.0) -> dict:
    """Calculates peak freq (wmax), gain(wmax), phase(wmax), grp-dly(wmax).

    Calculates the peak frequency, and the gain, phase and group delay at this
    frequency, for both over- and underdamped oscillators, and stores the
    results in a dictionary.

    The 'damp-type' key records the damp type for a particular cartesian pole,
    and the value of this record is the key to the array that stores the
    associated (wmax, gain, phase, group delay) for the pole.

    {
     'damp-type': [{"under"|"over"}]
     'under': [wmax, gain, phase, group delay],
     'over': [wmax, gain, phase, group delay],
    }

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.
    tau: float
        The impulse-response scale factor (default = 1.0).

    Returns
    -------
    dict
        Keys: ["damp-type", "over", "under"].
        Values: np.ndarrays.
    """

    # extract wn, theta
    wn, theta = extract_wn_and_theta(cartesian_poles)

    # build dict container
    n = wn.shape[0]
    d = {
        "damp-type": np.empty([n], dtype=object),
        "under": np.zeros((n, 4)),
        "over": np.zeros((n, 4)),
    }

    def eval_under_over(this_theta: float) -> str:
        if 0 <= np.abs(np.arctan(np.tan(this_theta))) < np.pi / 4:
            return "over"
        else:
            return "under"

    # apply and build iterator
    map_under_over = map(eval_under_over, theta)

    # iter over types
    for i, damp_type in enumerate(map_under_over):

        # record type
        d["damp-type"][i] = damp_type

        # locals
        wn_ = wn[i]
        th_ = theta[i]

        # overdamped calcs
        wmax_ = 0.0
        go_ = 1.0
        pho_ = 0.0
        gdo_ = 2.0 * np.cos(th_) / wn_ * tau

        d["over"][i, :] = np.array([wmax_, go_, pho_, gdo_])

        # underdamped calcs
        if damp_type == "under":

            wmax_ = wn_ / tau * np.sqrt(np.cos(np.pi - 2.0 * np.abs(th_)))
            gu_ = 1.0 / np.sin(2.0 * np.abs(th_))
            phu_ = -np.arctan(np.sqrt(np.power(np.tan(th_), 2) - 1.0))
            gdu_ = tau / (wn_ * np.cos(th_))

            d["under"][i, :] = np.array([wmax_, gu_, phu_, gdu_])

        else:

            d["under"][i, :] = np.array([np.nan, np.nan, np.nan, np.nan])

    # return
    return d


# noinspection SpellCheckingInspection
def autocorrelation_peak_and_stride_values(
    cartesian_poles: np.ndarray, tau=1.0
) -> dict:
    """Returns peak and stride values for array of poles, scaled by tau.

    The cartesian poles are converted to (wn, theta), where theta is zero
    on the negative real axis.

    For theta => 0,

                 wn sec(theta)
        kh(0) = --------------- .
                    4 tau

    The stride express depends on whether |theta| > 0 or theta == 0.
    For |theta| > 0,

                     sec(theta)
        xi_5pct = - ------------ log( 0.05 sin( |theta| ) ) tau,
                         wn

    while for theta == 0,

                   -1 - W_{-1}( -0.05 e^{-1} )
        xi_5pct = ----------------------------- tau
                               wn

    where W is the lambertw function, with branch index -1.

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.
    tau: float
        The impulse-response scale factor (default = 1.0).

    Returns
    -------
    dict
        Peak value (scaled by 1/tau) and 5% pct stride (scaled by tau)
    """

    # extract wn, theta
    wn, theta = extract_wn_and_theta(cartesian_poles)

    # eval kh0
    kh0 = wn / (4.0 * np.cos(theta))

    # eval xi_5pct
    xi_5pct = np.zeros_like(kh0)
    pct_5 = 0.05
    lam_arg = pct_5 * np.exp(-1.0)

    # it would be much better to build a lambda array and apply
    for i, th in enumerate(theta):

        if np.abs(th) > 0:
            xi_5pct[i] = -np.log(pct_5 * np.sin(np.abs(th))) / (
                wn[i] * np.cos(th)
            )
        else:
            xi_5pct[i] = (
                -1 - np.real(scipy.special.lambertw(-lam_arg, -1))
            ) / wn[i]

    # scale
    kh0 /= tau
    xi_5pct *= tau

    # compose dict and return
    return {"kh0": kh0, "xi_5pct": xi_5pct}


def cast_to_polar_form(cartesian_coords: np.ndarray) -> np.ndarray:
    """Converts complex values in Cartesian form to polar form.

    Polar angles are with respect to the positive real axis (as is standard).

    Parameters
    ----------
    cartesian_coords: np.ndarray
        (n, ) array of complex coordinates in cartesian form.

    Returns
    -------
    np.ndarray
        (n, 2) array of complex coordinates in polar form (abs, angle).
    """

    # allocate
    polar_coords = np.zeros((cartesian_coords.shape[0], 2))

    # convert
    polar_coords[:, 0] = np.abs(cartesian_coords)
    polar_coords[:, 1] = np.angle(cartesian_coords)

    # return
    return polar_coords


def cast_to_cartesian_form(polar_coords: np.ndarray) -> np.ndarray:
    """Converts polar representation of complex coordinates to Cartesian form.

    Polar angles are with respect to the positive real axis (as is standard).

    Parameters
    ----------
    polar_coords: np.ndarray
        (n, 2) array of complex coordinates in polar form (abs, angle).

    Returns
    -------
    np.ndarray
        (n, ) array of complex values in Cartesian form.
    """

    # convert and return
    return np.array(
        polar_coords[:, 0] * np.cos(polar_coords[:, 1])
        + 1j * polar_coords[:, 0] * np.sin(polar_coords[:, 1])
    )


def convert_reference_angle_to_negative_real_axis(
    polar_coords: np.ndarray,
) -> np.ndarray:
    r"""Converts the angle reference from positive to negative real axis.

    The angle of a complex coordinate is naturally taken with respect to the
    positive real axis. This angle is labeled phi. For the damped oscillator,
    it is convenient to reference angles the negative real axis (labeled
    theta) so that a "damping angle" of zero refers to a purely real pole
    location, instead of a "damping angle" of +/- pi.


     s-plane         ^  imag                s-plane         ^  imag
                     |                                      |
               x     +  pi                            x     +  pi
                \    |                                 \    |
                 \ <------ wn                           \ <------ wn
                  \  |                                   \--|-
                  /\ |                                    \ |  \  <---- phi
     theta --->  |  \|                                     \|   |
        -------+--\--+-------> real    <-->    -------+-----+---|----> real
              pr     |                               pr     |
                     |                                      |
                     |                                      |

    Applying this function twice recovers the original angle.

    Parameters
    ----------
    polar_coords: np.ndarray
        (n, 2) array of complex coordinates in polar form (abs, angle).

    Returns
    -------
    np.ndarray
        (n, 2) array where the angles (array[:, 1]) have been converted.

    """

    # convert and return
    polar_coords[:, 1] = np.pi - (polar_coords[:, 1] % (2 * np.pi))

    return polar_coords


def extract_wn_and_theta(cartesian_poles: np.ndarray):
    """Extracts natural frequency wn and inclination angle theta from coords.

    Parameters
    ----------
    cartesian_poles: np.ndarray
        Poles in cartesian coordinates.

    Returns
    -------
    tuple
        (wn, theta) as np.ndarrays.
    """

    # convert to polar form
    polar_poles = convert_reference_angle_to_negative_real_axis(
        cast_to_polar_form(cartesian_poles)
    )

    return polar_poles[:, 0], polar_poles[:, 1]



"""
-------------------------------------------------------------------------------

Analog multistage-box inline slope filter design:

Captures the design details for this filter. Design details are available
in Mathematica notebook mbox_inline_slope.nb. Wavenumbers are
calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import sys

from irides.design.analog import mbox_level as design_reference
from irides.tools import design_tools


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as variables in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "mbox-inline-slope"


def design_type() -> design_tools.FilterDesignType:
    """This is a slope filter"""

    return design_tools.FilterDesignType.SLOPE


def get_minimum_valid_filter_order(strict: bool = False) -> int:
    """Returns minimum valid filter order."""

    return design_tools.get_minimum_valid_filter_order(design_type(), strict)


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns array of valid filter orders based on strictness."""

    return design_tools.get_valid_filter_orders(design_type(), strict)


# noinspection SpellCheckingInspection,PyUnusedLocal
def impulse_response_t0_value(
    tau: float, filter_order: int, dt: float
) -> float:
    """
    Returns the h_m(t=0) value for inline slope filters of this type. Only
    orders [2..8] are provided.

    Parameters
    ----------
    dt: float
        Temporal discretization
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        Theoretic value of h_m(t=0).
    """

    # values calculated from Mathematica
    unit_t0_values = np.array(
        [
            1.0,  # m = 2
            0.0,  # m = 3
            0.0,  # m = 4
            0.0,  # m = 5
            0.0,  # m = 6
            0.0,  # m = 7
            0.0,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return unit_t0_values[index] / np.power(tau, 2)


def zero_crossing_time(tau: float, filter_order: int) -> float:
    r"""
    The zero-crossing time `tx` is when h_slope(tx) = 0. This is
    the same time when h_ref(tx) reaches its maximum.

      h_slope(t)
          ^       .
          |    ... ...
             ..       .
          ---|----|----0----|----------------------------> t
             0          .       ....
                         ... ...
                       ^    .
                       |
                      tx

    Only orders [2..8] are provided.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        Time tx where h_slope(tx) = 0.
    """

    # values calculated from Mathematica
    unit_crossing_times = np.array(
        [
            1.0,  # m = 2
            1.0,  # m = 3
            1.0,  # m = 4
            1.0,  # m = 5
            1.0,  # m = 6
            1.0,  # m = 7
            1.0,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return unit_crossing_times[index] * tau


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Note that the generated functions are not purely continuous time.
    Instead, they bridge continuous and discrete time by applying order-dt
    corrections to the moments. This is required because the CT series are
    in fact implemented in DT.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with `tau` and `dt` capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int) -> float:
        """Covers order [2..8] only."""
        return 0.0

    # noinspection PyUnusedLocal
    def m1_gen(order: int) -> float:
        return -1.0

    # noinspection PyUnusedLocal
    def m2_gen(order: int) -> float:
        return -2.0 * tau

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection, PyUnusedLocal
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int, dt=0.0
) -> dict:
    """
    System autocorrelation peak and stride values. Peak value is kh(0),
    and the stride comes from the trailing kh(xi_5pct) / kh(0) = -5%.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter (note: user is required to validate that
        the order is supported).
    dt: float
        [Optional] Temporal increment dt in h(t).

    Returns
    -------
    dict
        Peak value (scaled by 1/tau^3) and 5% pct stride (scaled by tau)
    """

    # values from Mathematica

    unit_sacf_values = {
        2: {"kh0": 2.0, "xi_5pct": 1.9000000000},
        3: {"kh0": 27.0 / 8.0, "xi_5pct": 1.5537113670},
        4: {"kh0": 16.0 / 3.0, "xi_5pct": 1.3393447220},
        5: {"kh0": 4375.0 / 576.0, "xi_5pct": 1.1953057840},
        6: {"kh0": 809.0 / 80.0, "xi_5pct": 1.0894011890},
        7: {"kh0": 1481417.0 / 115200.0, "xi_5pct": 1.0073948210},
        8: {"kh0": 224228.0 / 14175.0, "xi_5pct": 0.9414763172},
    }

    # fetch values for order and scale
    sacf_values = unit_sacf_values[filter_order]
    if filter_order == 2:
        sacf_values["kh0"] -= 2.0 * dt
    sacf_values["kh0"] /= np.power(tau, 3)
    sacf_values["xi_5pct"] *= tau

    # return
    return sacf_values


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        2: {"sti_5pct": 1.87194, "residual": +0.05000},
        3: {"sti_5pct": 3.31446, "residual": -0.05000},
        4: {"sti_5pct": 3.21686, "residual": -0.05000},
        5: {"sti_5pct": 2.95635, "residual": -0.05000},
        6: {"sti_5pct": 2.73096, "residual": -0.05000},
        7: {"sti_5pct": 2.55439, "residual": -0.05000},
        8: {"sti_5pct": 2.41530, "residual": -0.05000},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return 0.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return np.pi / 2


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return design_reference.group_delay_at_dc(tau)


# noinspection SpellCheckingInspection
def frequency_of_peak_gain(tau: float, filter_order: int) -> float:
    """Returns the radial frequency at which the gain peaks."""

    # values calculated from Mathematica
    unit_frequency_of_peak_gain = np.array(
        [
            2.331122356,  # m = 2
            2.902207933,  # m = 3
            3.378923335,  # m = 4
            3.796538379,  # m = 5
            4.172699669,  # m = 6
            4.517719693,  # m = 7
            4.838240217,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return unit_frequency_of_peak_gain[index] / tau


# noinspection SpellCheckingInspection
def gain_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns gain(w_peak), the peak gain."""

    # values calculated from Mathematica
    unit_gain_at_peak_frequency = np.array(
        [
            1.449222708,  # m = 2
            1.789626825,  # m = 3
            2.075053083,  # m = 4
            2.325758038,  # m = 5
            2.551973194,  # m = 6
            2.759718734,  # m = 7
            2.952891372,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return unit_gain_at_peak_frequency[index] / tau


# noinspection SpellCheckingInspection,PyUnusedLocal
def phase_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns angle H_delta(s) at the peak frequency."""

    # values calculated from Mathematica
    phase_at_peak_frequency_values = np.array(
        [
            -0.760326029,  # m = 2
            -1.331411606,  # m = 3
            -1.808127008,  # m = 4
            -2.225742052,  # m = 5
            -2.601903342,  # m = 6
            -2.946923366,  # m = 7
            -3.267443890,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return phase_at_peak_frequency_values[index]


# noinspection SpellCheckingInspection
def group_delay_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns the group delay at the peak frequency."""

    # values calculated from Mathematica
    unit_group_delay_at_peak_frequency = np.array(
        [
            1.0,  # m = 2
            1.0,  # m = 3
            1.0,  # m = 4
            1.0,  # m = 5
            1.0,  # m = 6
            1.0,  # m = 7
            1.0,  # m = 8
        ]
    )

    # return
    index = filter_order - get_minimum_valid_filter_order()
    return unit_group_delay_at_peak_frequency[index] * tau


# noinspection SpellCheckingInspection,PyPep8Naming
def wireframe(tau: float, filter_order: int):
    """Wireframe-timepoint-pair for this slope filter.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_slope_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        2: 0.80573771,
        3: 0.83797329,
        4: 0.83308778,
        5: 0.82091467,
        6: 0.79784648,
        7: 0.76948133,
        8: 0.74256534,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        2: {"TW": 0.57735, "BW": 741.036, "UCP": 427.838},
        3: {"TW": 0.447214, "BW": 3.67423, "UCP": 1.64317},
        4: {"TW": 0.377964, "BW": 4.0, "UCP": 1.51186},
        5: {"TW": 0.333333, "BW": 4.51189, "UCP": 1.50396},
        6: {"TW": 0.301511, "BW": 4.98304, "UCP": 1.50244},
        7: {"TW": 0.27735, "BW": 5.4146, "UCP": 1.50174},
        8: {"TW": 0.258199, "BW": 5.81455, "UCP": 1.50131},
    }

    return uncertainty_products[filter_order]



"""
-------------------------------------------------------------------------------

Digital Bessel filter design:

The fundamental file is analog.bessel_level.py because KT-Bessel filters are
designed in continuous time. Yet, for design-level parameters necessary
for digital implementation, this file captures these additional values.
Bessel filters are designed in continuous time

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import bessel_level as dsgn_alg_bssl
from irides.tools import design_tools
from irides.tools import digital_design_tools as dd_tools
from irides.tools import analog_to_digital_conversion_tools as a2d_tools

"""
-------------------------------------------------------------------------------

Filter-specific definitions

-------------------------------------------------------------------------------
"""


# noinspection SpellCheckingInspection
def designs(filter_order: int) -> dict:
    """Design points related to digital Bessel filter realizations."""

    # designs
    design = {}

    design[1] = {
        "tau_minimum": 1.4426950408889634,
        "mu_minimum": 1.0,
        "scaled_bandwidth_pct": 0.0,
    }

    design[2] = {
        "tau_minimum": 2.3950324087595174,
        "mu_minimum": 1.4994039481201957,
        "scaled_bandwidth_pct": 0.23019684134769067,
    }

    design[3] = {
        "tau_minimum": 3.3487313481270853,
        "mu_minimum": 1.9983665156513852,
        "scaled_bandwidth_pct": 0.33352141188877066,
    }

    design[4] = {
        "tau_minimum": 4.302913367060106,
        "mu_minimum": 2.4971877143015915,
        "scaled_bandwidth_pct": 0.39316730888500784,
    }

    design[5] = {
        "tau_minimum": 5.25732894539093,
        "mu_minimum": 2.9959541441532784,
        "scaled_bandwidth_pct": 0.4324218290845279,
    }

    design[6] = {
        "tau_minimum": 6.2118775638638635,
        "mu_minimum": 3.494696994129047,
        "scaled_bandwidth_pct": 0.46042833324907484,
    }

    design[7] = {
        "tau_minimum": 7.1665102003888395,
        "mu_minimum": 3.9934293398661236,
        "scaled_bandwidth_pct": 0.48153438243678787,
    }

    design[8] = {
        "tau_minimum": 8.121199813729849,
        "mu_minimum": 4.492157233932513,
        "scaled_bandwidth_pct": 0.49808226091458135,
    }

    # check and return
    return design[filter_order]


"""
-------------------------------------------------------------------------------

Standard features

-------------------------------------------------------------------------------
"""


def design_id() -> str:
    """Design id of this filter."""

    return "digital-{0}".format(dsgn_alg_bssl.design_id())


def design_type() -> design_tools.FilterDesignType:
    """This is a level filter"""

    return dsgn_alg_bssl.design_type()


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design."""

    return dsgn_alg_bssl.get_valid_filter_orders(strict)


def minimum_mu_value(filter_order: int) -> float:
    """Returns the minimum mu value for a given filter order"""

    return designs(filter_order)["mu_minimum"]


# noinspection PyPep8Naming
def gain_adjustment(mu: float, filter_order: int, T: float = 1) -> float:
    """Returns the gain adjustment so that the level-filter's gain is unity.

    The gain adjustment for digital poles pd_k is

        gain_adj = (1 - pd_1) (1 - pd_2) ... (1 - pd_m),

    and the real part is taken to annihilate the zero-valued imaginary part.
    """

    # fetch zplane poles
    zplane_poles = a2d_tools.zplane_poles_from_analog_design(
        dsgn_alg_bssl, mu, filter_order, T
    )

    # compute and return
    return np.real(np.exp(np.sum(np.log(1.0 - zplane_poles))))


# noinspection PyPep8Naming
def initial_hn_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated values of h[0], h[1], and h[2].

    From a general single-pole product expansion of H(z), we find that

        h[0] = gain_adj
        h[1] = gain_adj * sum( digital_poles )
        h[2] = gain_adj * (sum( digital_poles )^2 - residual)

    where, as a 4th-order example to show the pattern,

        residual = p1 (p2 + p3 + p4) + p2 (p3 + p4) + p3 * p4.
    """

    # prerequisites
    zplane_poles = a2d_tools.zplane_poles_from_analog_design(
        dsgn_alg_bssl, mu, filter_order
    )
    # Hz = construct_Hz(mu, filter_order)
    Hz = dd_tools.construct_transfer_function(mu, filter_order, dsgn_alg_bssl)

    # calculate h[0]
    h0 = Hz.initial_value()

    # calculate h[1]
    zplane_poles_sum = np.sum(zplane_poles)
    h1 = h0 * np.real(zplane_poles_sum)

    # calculate h[2]
    residual = np.real(
        np.sum(
            [
                zplane_poles[i] * np.sum(zplane_poles[i + 1 :])
                for i in range(filter_order - 1)
            ]
        )
    )
    h2 = h0 * np.real(np.power(zplane_poles_sum, 2) - residual)

    # pack and return as an array
    return np.array([h0, h1, h2])


# noinspection PyPep8Naming
def moment_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns moment values (M0, M1, M2) computed from H(z).

    Since the analytic derivatives of H(z) are complicated, it is simpler
    to numerically compute its approximate derivatives. To improve accuracy,
    higher-order finite differences are taken.
    """

    # build an H(z) object
    # Hz = construct_Hz(mu, filter_order)
    Hz = dd_tools.construct_transfer_function(mu, filter_order, dsgn_alg_bssl)

    # setup to H'(1) and H''(1)
    scale = 1e-4
    eps = scale / mu
    zvals = 1.0 + eps * np.arange(-3, 3 + 1)
    Hz_pts = Hz.value(zvals)

    # approximate from differences

    # these coefs from
    #   https://en.wikipedia.org/wiki/Finite_difference_coefficient
    zero_diff_coefs = np.array([0, 0, 0, 1, 0, 0, 0])
    first_diff_coefs = (
        np.array([-1 / 60, 3 / 20, -3 / 4, 0, 3 / 4, -3 / 20, 1 / 60]) / eps
    )
    second_diff_coefs = np.array(
        [1 / 90, -3 / 20, 3 / 2, -49 / 18, 3 / 2, -3 / 20, 1 / 90]
    ) / np.power(eps, 2)

    H = np.sum(zero_diff_coefs * Hz_pts)
    Hp = np.sum(first_diff_coefs * Hz_pts)
    Hpp = np.sum(second_diff_coefs * Hz_pts)

    # M0 and M1 and M2
    M0 = H
    M1 = -Hp
    M2 = Hp + Hpp

    # pack and return
    return np.array([M0, M1, M2])


def full_width_value(mu: float, filter_order: int) -> float:
    """Returns the numerically approximated FW value"""

    moments = moment_values(mu, filter_order)

    return 2.0 * np.sqrt(moments[2] - np.power(moments[1], 2))


def analog_tau_value(mu: float, filter_order: int) -> float:
    """Returns `tau` associated with (`mu`, `filter_order`)"""

    # compute and return
    splane_poles = dsgn_alg_bssl.designs(filter_order)["poles"]

    return a2d_tools.solve_for_tau_from_mu_and_splane_poles(mu, splane_poles)



#!/usr/bin/env python

# Original from
#  https://raw.githubusercontent.com/ionelmc/python-nameless/purepython/setup.py

# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    # id
    name="irides",
    url="https://gitlab.com/irides-dev/irides",
    # key meta
    version="1.0.2",
    license="Apache License 2.0",
    description="""Irides provides references implementations for select linear time-invariant filters.""",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    # authorship and related
    author="Jay Damask",
    author_email="jaydamask@buell-lane-press.co",
    # project meta
    project_urls={
        "Documentation": "https://gitlab.com/irides-dev/irides/",
        "Changelog": "https://gitlab.com/irides-dev/irides/",
        "Issue Tracker": "https://gitlab.com/irides-dev/irides/issues",
    },
    keywords=["multivariate statistics"],
    classifiers=[
        # ref to https://pypi.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
    ],
    # packaging details
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    # requirements and dependencies
    python_requires=">=3.7",
    install_requires=["numpy>=1.23"],
    extras_require={},
)



"""
-------------------------------------------------------------------------------

Discrete-time unit-step operator.

-------------------------------------------------------------------------------
"""

import numpy as np
from irides.resources.containers.discrete_sequence import DiscreteSequence
from irides.tools.digital_design_tools import FrequencyBand


# -----------------------------------------------------------------------------
# design


def delay(n_end: int = np.inf) -> float:
    """Returns the average delay of the operator (applies only to LOW band)"""

    return n_end / 2


# -----------------------------------------------------------------------------
# filter signature


def generate_impulse_response(
    n_end: int, frequency_band: FrequencyBand = FrequencyBand.LOW
) -> DiscreteSequence:
    """Returns a discrete sequence of (+/-1)^n between [0, n_end).

    Parameters
    ----------
    n_end: int
        The value of the highest index (exclusive of this point).
    frequency_band: FrequencyBand
        LOW frequency -> 1^n, HIGH frequency -> (-1)^n

    Returns
    -------
    DiscreteSequence
        A preloaded discrete sequence.
    """

    # setup a discrete sequence
    ds = DiscreteSequence(0, n_end)

    # set weights
    signed_one = 1 if frequency_band == FrequencyBand.LOW else -1
    ds.v_axis = np.power(signed_one, ds.n_axis)

    # return
    return ds



"""
-------------------------------------------------------------------------------

Implements an ideal delay in a discretized continuous time.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.tools import impulse_response_builder_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(tau: float, dt: float) -> np.ndarray:
    """Creates an ideal delay time series that can be used for convolution.

    The discretization naturally replaces a Dirac delta with a Kronecker delta
    with a weight of 1 / dt.

    Parameters
    ----------
    tau: float
        Timepoint of the Dirac delta (tau >= 0).
    dt: float
        Time increment.

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h_delay].
    """

    # determine (t_start, t_end)
    t_start = 0.0
    t_end = np.ceil(tau / dt) * dt

    # init time series
    (
        time_series,
        i_t0_plus,
    ) = impulse_response_builder_tools.make_impulse_response_template(
        t_start, t_end, dt
    )

    # locate the `tau` timepoint
    tau_sample = int(np.round(tau / dt))

    # set the delta
    time_series[i_t0_plus + tau_sample, 1] = 1.0 / dt

    # return
    return time_series



"""
-------------------------------------------------------------------------------

Core enumerations for the irides package.

-------------------------------------------------------------------------------
"""

from enum import Enum


class FilterType(Enum):
    """Universe of filter types."""

    MBOX = 0
    PEMA = 1
    BESSEL = 2
    UNKNOWN = 3


# noinspection SpellCheckingInspection,PyPep8Naming
class FilterDesignType(Enum):
    """Universe of labels that represent implemented filter-operation types."""

    LEVEL = 0
    SLOPE = 1
    CURVE = 2
    UNKNOWN = 3


class TestSignalType(Enum):
    """Universe of test-signal types."""

    IMPULSE = 0
    STEP = 1
    RAMP = 2
    PARABOLA = 3
    ALTERNATING = 4
    WHITE_NOISE = 5
    UNKNOWN = 6



"""
-------------------------------------------------------------------------------

Discrete-time first-difference operator.

-------------------------------------------------------------------------------
"""

import numpy as np
from irides.resources.containers.discrete_sequence import DiscreteSequence


# -----------------------------------------------------------------------------
# design


def delay() -> float:
    """Returns the average delay of the operator"""

    return 0.5


# -----------------------------------------------------------------------------
# filter signature


def generate_impulse_response() -> DiscreteSequence:
    """Returns a discrete sequence two samples long with weights [+1, -1].

    Returns
    -------
    DiscreteSequence
        A preloaded discrete sequence.
    """

    # setup a discrete sequence
    ds = DiscreteSequence(0, 2)

    # set weights for the zeroth and first samples
    ds.v_axis = np.array([+1.0, -1.0])

    # return
    return ds



"""
-------------------------------------------------------------------------------

Analog unit-step filter signature:

Implements the temporal and spectral signatures of a continuous-time
unit-step function u(t).

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.tools import transfer_function_tools


def generate_impulse_response(tstart, tend, dt):
    r"""
    Returns an np array, first column is the (discrete) time axis,
    second column is the impulse response.

    The impulse response of a unit step is

                /  0, t <= 0
        u(t) = <
                \  1, t >= 0-

    where 0- captures any transition at t = 0.

    input:  /tstart/  start time
            /tend/    end time
            /dt/      time increment

    output: /panel/   np array, shape (npts, 2)

    """

    # calculate the number of points to eval on
    npts = int(np.ceil((tend - tstart) / dt))

    # setup the panel return
    panel = np.ndarray([npts, 2], dtype=float)

    # compute the time axis
    panel[:, 0] = [v * dt + tstart for v in range(npts)]

    # compute the impulse response
    panel[:, 1] = [0.0 if t < 0.0 else 1.0 for t in panel[:, 0]]

    # return
    return panel


def generate_transfer_function(fstart, fend, df):
    """
    Returns an np array, first column is the (cyclic) frequency axis,
    second column is the complex-valued transfer function.

    The defining transfer function is

                1
        H(s) = --- .
                s

    In general, s = omega + j w, a complex value. Here, however, omega = 0,
    so /s/ is replaced by /j w/.

    Moreover, the api is for cyclic frequency /f/, measured in cycles per sec,
    but radial frequency /w/, measured in radians per sec, is used internally.

    input:  /fstart/        start frequency, cycles per second
            /fend/          end frequency
            /df/            frequency increment

    output: /panel/         np array, shape (npts, 2)

    """

    # calculate the number of points to eval on
    npts = int(np.ceil((fend - fstart) / df))

    # setup the panel return
    panel = np.ndarray([npts, 2], dtype=complex)

    # compute the frequency axis
    panel[:, 0] = [v * df + fstart for v in range(npts)]

    # compute the complex transfer function along the /jw/ axis
    jw = 1j * 2.0 * np.pi * panel[:, 0]

    with np.errstate(divide="ignore", invalid="ignore"):
        panel[:, 1] = 1.0 / jw

    # return
    return panel


def generate_gain_spectrum(fstart, fend, df):
    """
    Generates the gain spectrum of a unit-step filter.

    input:  /fstart/        start frequency, cycles per second
            /fend/          end frequency
            /df/            frequency increment

    output: /panel/         np array, shape (npts, 2)
                                column 1: [fstart, fend]
                                column 2: [gain spectrum]
    """

    # get the transfer function panel
    xfer_fcxn_panel = generate_transfer_function(fstart, fend, df)

    # convert to a gain spectrum
    spectral_panel = transfer_function_tools.calculate_gain_spectrum(
        xfer_fcxn_panel
    )

    # return
    return spectral_panel


def generate_phase_spectrum(fstart, fend, df):
    """
    Generates the phase spectrum of a unit-step filter.

    input:  /fstart/        start frequency, cycles per second
            /fend/          end frequency
            /df/            frequency increment

    output: /panel/         np array, shape (npts, 2)
                                column 1: [fstart, fend]
                                column 2: [phase spectrum]
    """

    # get the transfer function panel
    xfer_fcxn_panel = generate_transfer_function(fstart, fend, df)

    # convert to a phase spectrum
    xfer_fcxn_panel = transfer_function_tools.perturb_dc_frequency(
        xfer_fcxn_panel, df
    )
    spectral_panel = transfer_function_tools.calculate_phase_spectrum(
        xfer_fcxn_panel
    )

    # return
    return spectral_panel


def generate_group_delay_spectrum(fstart, fend, df):
    """
    Generates the group-delay spectrum of a unit-step filter.

    input:  /fstart/        start frequency, cycles per second
            /fend/          end frequency
            /df/            frequency increment

    output: /panel/         np array, shape (npts, 2)
                                column 1: [fstart, fend]
                                column 2: [group-delay spectrum]
    """

    # get the transfer function panel with an extra row at fend + df
    xfer_fcxn_panel = generate_transfer_function(fstart, fend + df, df)

    # compute the group-delay spectrum
    spectral_panel = transfer_function_tools.calculate_group_delay_spectrum(
        xfer_fcxn_panel, df
    )

    # return
    return spectral_panel[:-1, :]
